/*
 * Copyright (C) 2014, United States Government, as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All rights reserved.
 *
 * The Java Pathfinder core (jpf-core) platform is licensed under the
 * Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0.
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package gov.nasa.jpf.listener;

import gov.nasa.jpf.Config;
import gov.nasa.jpf.JPF;
import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.search.Search;
import gov.nasa.jpf.jvm.bytecode.*;
import gov.nasa.jpf.vm.*;
import gov.nasa.jpf.vm.bytecode.LocalVariableInstruction;
import gov.nasa.jpf.vm.bytecode.ReadInstruction;
import gov.nasa.jpf.vm.bytecode.StoreInstruction;
import gov.nasa.jpf.vm.bytecode.WriteInstruction;
import gov.nasa.jpf.vm.choice.IntIntervalGenerator;

import java.io.PrintWriter;

import java.util.*;

// TODO: Fix for Groovy's model-checking
// TODO: This is a listener created to detect device conflicts and global variable conflicts
/**
 * Simple listener tool to track remove unwanted ChoiceGenerators that are registered
 * while an existing ChoiceGenerator is being explored. Multiple ChoiceGenerators of the same
 * type could be registered spuriously during a while(true) loop.
 */
public class StateReducer extends ListenerAdapter {

  private boolean debugMode;
  private boolean stateReductionMode;
  private final PrintWriter out;
  private String detail;
  private int depth;
  private int id;
  private Transition transition;
  private HashSet<String> conflictSet;
  private HashSet<String> appSet;
  // Holds the event value numbers that execute write instructions
  private HashSet<Integer> writeEventNumberSet;
  // Holds values that have appeared during CG advances
  private HashSet<Integer> cgChoiceSet;
  private Integer choiceUpperBound;
  private boolean isInitialized;
  private int currentChoice;
  // Records how many times BooleanChoiceGenerator has been advanced
  private int bCGIsAdvanced;
  // Records how many times we have looped through all the event numbers (IntIntervalGenerator)
  private int iiCGIsLooped;

  // Constants to determine when to start reducing states
  private final int BOOLEAN_CG_ADVANCED_THRESHOLD = 2;
  private final int INTEGER_CG_ADVANCED_THRESHOLD = 50;

  public StateReducer (Config config) {
    debugMode = config.getBoolean("debug_state_transition", false);
    stateReductionMode = config.getBoolean("activate_state_reduction", true);
    if (debugMode) {
      out = new PrintWriter(System.out, true);
    } else {
      out = null;
    }
    detail = null;
    depth = 0;
    id = 0;
    transition = null;
    cgChoiceSet = new HashSet<>();
    choiceUpperBound = 0;
    isInitialized = false;
    conflictSet = new HashSet<>();
    String[] conflictVars = config.getStringArray("variables");
    // We are not tracking anything if it is null
    if (conflictVars != null) {
      for (String var : conflictVars) {
        conflictSet.add(var);
      }
    }
    appSet = new HashSet<>();
    String[] apps = config.getStringArray("apps");
    // We are not tracking anything if it is null
    if (apps != null) {
      for (String var : apps) {
        appSet.add(var);
      }
    }
    currentChoice = 0;
    writeEventNumberSet = new HashSet<>();
    bCGIsAdvanced = 0;
    iiCGIsLooped = 0;
  }

  @Override
  public void stateRestored(Search search) {
    if (debugMode) {
      id = search.getStateId();
      depth = search.getDepth();
      transition = search.getTransition();
      detail = null;
      out.println("\n==> DEBUG: The state is restored to state with id: " + id + " -- Transition: " + transition +
              " and depth: " + depth + "\n");
    }
  }

  //--- the ones we are interested in
  @Override
  public void searchStarted(Search search) {
    if (debugMode) {
      out.println("\n==> DEBUG: ----------------------------------- search started" + "\n");
    }
  }

  @Override
  public void choiceGeneratorSet (VM vm, ChoiceGenerator<?> newCG) {
    if (stateReductionMode) {
      // Initialize with necessary information from the CG
      if (newCG instanceof IntIntervalGenerator) {
        IntIntervalGenerator iigCG = (IntIntervalGenerator) newCG;
        // Check if CG has been initialized, otherwise initialize it
        if (!isInitialized) {
          Integer[] choices = iigCG.getChoices();
          // Get the upper bound from the last element of the choices
          choiceUpperBound = choices[choices.length - 1];
          isInitialized = true;
        } else {
          newCG.setDone();
        }
      }
    }
  }

  @Override
  public void choiceGeneratorAdvanced (VM vm, ChoiceGenerator<?> currentCG) {
    if (stateReductionMode) {
      // Check for BooleanChoiceGenerator
      if (currentCG instanceof BooleanChoiceGenerator) {
        bCGIsAdvanced++;
      }
      // Check every choice generated and make sure that all the available choices
      // are chosen first before repeating the same choice of value twice!
      if (currentCG instanceof IntIntervalGenerator) {
        IntIntervalGenerator iigCG = (IntIntervalGenerator) currentCG;
        Integer nextChoice = iigCG.getNextChoice();
        if (!cgChoiceSet.contains(nextChoice)) {
          cgChoiceSet.add(nextChoice);
        }
        // Allow reinitialization after an upper bound is hit
        // This means all available choices have been explored once during this iteration
        if (cgChoiceSet.contains(choiceUpperBound)) {
          isInitialized = false;
          cgChoiceSet.clear();
          iiCGIsLooped++;
        }
        // Record the current choice
        currentChoice = iigCG.getNextChoice();
        // Start reducing (advancing) if the thresholds are passed
        if (iiCGIsLooped >= INTEGER_CG_ADVANCED_THRESHOLD &&
            bCGIsAdvanced >= BOOLEAN_CG_ADVANCED_THRESHOLD) {
          if (!writeEventNumberSet.contains(currentChoice)) {
            iigCG.advance();
          }
        }
      }
    }
  }

  @Override
  public void stateAdvanced(Search search) {
    if (debugMode) {
      id = search.getStateId();
      depth = search.getDepth();
      transition = search.getTransition();
      if (search.isNewState()) {
        detail = "new";
      } else {
        detail = "visited";
      }

      if (search.isEndState()) {
        out.println("\n==> DEBUG: This is the last state!\n");
        detail += " end";
      }
      out.println("\n==> DEBUG: The state is forwarded to state with id: " + id + " with depth: " + depth +
              " which is " + detail + " Transition: " + transition + "\n");
    }
  }

  @Override
  public void stateBacktracked(Search search) {
    if (debugMode) {
      id = search.getStateId();
      depth = search.getDepth();
      transition = search.getTransition();
      detail = null;

      out.println("\n==> DEBUG: The state is backtracked to state with id: " + id + " -- Transition: " + transition +
              " and depth: " + depth + "\n");
    }
  }

  @Override
  public void searchFinished(Search search) {
    if (debugMode) {
      out.println("\n==> DEBUG: ----------------------------------- search finished" + "\n");
    }
  }

  @Override
  public void instructionExecuted(VM vm, ThreadInfo ti, Instruction nextInsn, Instruction executedInsn) {
    // CASE #1: Detecting variable write-after-write conflict
    if (executedInsn instanceof WriteInstruction) {
      // Check for write-after-write conflict
      String varId = ((WriteInstruction) executedInsn).getFieldInfo().getFullName();
      for(String var : conflictSet) {
        if (varId.contains(var)) {
          String writer = getWriter(ti.getStack());
          // Just return if the writer is not one of the listed apps in the .jpf file
          if (writer == null)
            return;

          // Stores event number that executes WriteInstruction
          if (!writeEventNumberSet.contains(currentChoice)) {
            writeEventNumberSet.add(currentChoice);
          }
        }
      }
    }
  }

  // Find the variable writer
  // It should be one of the apps listed in the .jpf file
  private String getWriter(List<StackFrame> sfList) {
    // Start looking from the top of the stack backward
    for(int i=sfList.size()-1; i>=0; i--) {
      MethodInfo mi = sfList.get(i).getMethodInfo();
      if(!mi.isJPFInternal()) {
        String method = mi.getStackTraceName();
        // Check against the apps in the appSet
        for(String app : appSet) {
          // There is only one writer at a time but we need to always
          // check all the potential writers in the list
          if (method.contains(app)) {
            return app;
          }
        }
      }
    }

    return null;
  }
}